# Copyright (c) 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

import logging
import math
import os
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Optional, Union

import librosa
import mlx.core as mx
import mlx.nn as nn
import numpy as np

from mlx_audio.tts.models.base import GenerationResult

from .models.s3gen import S3GEN_SIL, S3GEN_SR, S3Gen
from .models.s3tokenizer import S3TokenizerV2, log_mel_spectrogram
from .models.t3 import T3, T3Cond, T3Config
from .models.voice_encoder import VoiceEncoder

logger = logging.getLogger(__name__)

# Constants
S3_SR = 16000  # S3Tokenizer sample rate
REPO_ID = "ResembleAI/chatterbox-turbo"


def punc_norm(text: str) -> str:
    """
    Quick cleanup func for punctuation from LLMs or
    containing chars not seen often in the dataset.
    """
    if len(text) == 0:
        return "You need to add some text for me to talk."

    # Capitalise first letter
    if text[0].islower():
        text = text[0].upper() + text[1:]

    # Remove multiple space chars
    text = " ".join(text.split())

    # Replace uncommon/llm punc
    punc_to_replace = [
        ("…", ", "),
        (":", ","),
        ("—", "-"),
        ("–", "-"),
        (" ,", ","),
        (
            """, '"'),
        (""",
            '"',
        ),
        ("'", "'"),
        ("'", "'"),
    ]
    for old_char_sequence, new_char in punc_to_replace:
        text = text.replace(old_char_sequence, new_char)

    # Add full stop if no ending punc
    text = text.rstrip(" ")
    sentence_enders = {".", "!", "?", "-", ","}
    if not any(text.endswith(p) for p in sentence_enders):
        text += "."

    return text


@dataclass
class Conditionals:
    """
    Conditionals for T3 and S3Gen.
    """

    t3: T3Cond
    gen: dict

    def save(self, fpath: Path):
        """Save conditionals to file."""
        import pickle

        with open(fpath, "wb") as f:
            pickle.dump({"t3": self.t3, "gen": self.gen}, f)

    @classmethod
    def load(cls, fpath: Path) -> "Conditionals":
        """Load conditionals from file."""
        import pickle

        with open(fpath, "rb") as f:
            data = pickle.load(f)
        return cls(data["t3"], data["gen"])


class ChatterboxTurboTTS(nn.Module):
    """
    MLX implementation of Chatterbox Turbo TTS.
    Optimized for Apple Silicon.
    """

    ENC_COND_LEN = 15 * S3_SR  # 15 seconds for encoder conditioning
    DEC_COND_LEN = 10 * S3GEN_SR  # 10 seconds for decoder conditioning

    def __init__(
        self,
        config_or_t3: Union[dict, T3] = None,
        s3gen: S3Gen = None,
        ve: VoiceEncoder = None,
        tokenizer=None,  # HuggingFace tokenizer
        s3tokenizer: S3TokenizerV2 = None,  # Speech tokenizer for conditioning
        conds: Optional[Conditionals] = None,
        local_path: Optional[str] = None,
    ):
        super().__init__()
        self.sr = S3GEN_SR  # Output sample rate

        # Check if first argument is a config dict (from load_model)
        if config_or_t3 is None or isinstance(config_or_t3, dict):
            # Initialize from config
            self.config = config_or_t3 or {}
            hp = T3Config.turbo()
            self.t3 = T3(hp)
            self.s3gen = S3Gen(meanflow=True)
            self.ve = VoiceEncoder()
        else:
            # Initialize with individual components
            self.config = {}
            self.t3 = config_or_t3
            self.s3gen = s3gen if s3gen is not None else S3Gen(meanflow=True)
            self.ve = ve if ve is not None else VoiceEncoder()

        self.tokenizer = tokenizer
        # S3 speech tokenizer for reference audio tokenization
        self._s3tokenizer = s3tokenizer or S3TokenizerV2("speech_tokenizer_v2_25hz")
        self.conds = conds
        self.local_path = local_path

    @property
    def sample_rate(self) -> int:
        """Output sample rate."""
        return self.sr

    def sanitize(self, weights: dict) -> dict:
        """
        Sanitize PyTorch weights for MLX.

        Routes weights to the appropriate component based on prefix:
        - ve.* -> VoiceEncoder weights
        - t3.* -> T3 model weights
        - s3gen.* -> S3Gen weights

        Args:
            weights: Dictionary of weight name -> array

        Returns:
            Sanitized weights dictionary
        """
        new_weights = {}

        # Separate weights by component prefix
        ve_weights = {}
        t3_weights = {}
        s3gen_weights = {}
        other_weights = {}

        for key, value in weights.items():
            if key.startswith("ve."):
                ve_weights[key[3:]] = value
            elif key.startswith("t3."):
                t3_weights[key[3:]] = value
            elif key.startswith("s3gen."):
                s3gen_weights[key[6:]] = value
            else:
                other_weights[key] = value

        # Sanitize each component's weights if they have sanitize methods
        if ve_weights:
            if hasattr(self.ve, "sanitize"):
                ve_sanitized = self.ve.sanitize(ve_weights)
            else:
                ve_sanitized = ve_weights
            for k, v in ve_sanitized.items():
                new_weights[f"ve.{k}"] = v

        if t3_weights:
            if hasattr(self.t3, "sanitize"):
                t3_sanitized = self.t3.sanitize(t3_weights)
            else:
                t3_sanitized = t3_weights
            for k, v in t3_sanitized.items():
                new_weights[f"t3.{k}"] = v

        if s3gen_weights:
            if hasattr(self.s3gen, "sanitize"):
                s3gen_sanitized = self.s3gen.sanitize(s3gen_weights)
            else:
                s3gen_sanitized = s3gen_weights
            for k, v in s3gen_sanitized.items():
                new_weights[f"s3gen.{k}"] = v

        # Add other weights as-is
        new_weights.update(other_weights)

        return new_weights

    def load_weights(self, weights, strict: bool = True):
        """
        Load weights into the model.

        Uses strict=False by default for components because ChatterboxTurbo has
        several non-checkpoint parameters that are generated during initialization.

        Args:
            weights: List of (key, value) tuples or dict
            strict: If False, ignore missing/extra keys. Default True for
                    compatibility with utils.load_model().
        """
        if isinstance(weights, dict):
            weights = list(weights.items())

        # Split weights by component prefix
        ve_weights = []
        t3_weights = []
        s3gen_weights = []
        other_weights = []

        for k, v in weights:
            if k.startswith("ve."):
                ve_weights.append((k[3:], v))
            elif k.startswith("t3."):
                t3_weights.append((k[3:], v))
            elif k.startswith("s3gen."):
                s3gen_weights.append((k[6:], v))
            elif k.startswith("gen."):
                # Skip gen.* keys - these are conditionals, not model weights
                continue
            else:
                other_weights.append((k, v))

        # Load each component with strict=False to handle non-checkpoint params
        if ve_weights:
            logger.info(f"Loading {len(ve_weights)} VE weights")
            self.ve.load_weights(ve_weights, strict=False)

        if t3_weights:
            logger.info(f"Loading {len(t3_weights)} T3 weights")
            self.t3.load_weights(t3_weights, strict=False)

        if s3gen_weights:
            logger.info(f"Loading {len(s3gen_weights)} S3Gen weights")
            self.s3gen.load_weights(s3gen_weights, strict=False)

        # Warn about unrecognized weights
        if other_weights and strict:
            unrecognized = [k for k, _ in other_weights]
            logger.warning(f"Unrecognized weight keys: {unrecognized}")

        # Evaluate all parameters
        mx.eval(
            self.ve.parameters(),
            self.t3.parameters(),
            self.s3gen.parameters(),
        )
        logger.info("Weights loaded successfully")

    @staticmethod
    def post_load_hook(
        model: "ChatterboxTurboTTS", model_path: Path
    ) -> "ChatterboxTurboTTS":
        """
        Post-load hook called by load_model to initialize tokenizer and conditionals.

        Args:
            model: The loaded model instance
            model_path: Path to the model directory

        Returns:
            The model with tokenizer and conditionals initialized
        """
        model.local_path = str(model_path)

        # Load text tokenizer
        try:
            from transformers import AutoTokenizer

            tokenizer = AutoTokenizer.from_pretrained(str(model_path))
            if tokenizer.pad_token is None:
                tokenizer.pad_token = tokenizer.eos_token
            model.tokenizer = tokenizer
            logger.info("Loaded text tokenizer")
        except Exception as e:
            logger.warning(f"Could not load text tokenizer: {e}")

        # Load S3 speech tokenizer weights
        try:
            from huggingface_hub import hf_hub_download

            s3tok_repo = "mlx-community/S3TokenizerV2"
            s3tok_weights_path = hf_hub_download(
                repo_id=s3tok_repo,
                filename="model.safetensors",
            )
            s3tok_weights = mx.load(s3tok_weights_path)
            if hasattr(model._s3tokenizer, "sanitize"):
                s3tok_weights = model._s3tokenizer.sanitize(s3tok_weights)
            model._s3tokenizer.load_weights(list(s3tok_weights.items()), strict=False)
            logger.info("Loaded S3 speech tokenizer weights")
        except Exception as e:
            logger.warning(f"Could not load S3 speech tokenizer: {e}")

        # Load pre-computed conditionals (prefer safetensors, fallback to .pt)
        builtin_voice_safetensors = model_path / "conds.safetensors"

        if builtin_voice_safetensors.exists():
            try:
                conds_data = mx.load(str(builtin_voice_safetensors))

                speaker_emb = conds_data.get("t3.speaker_emb")
                if speaker_emb is None:
                    speaker_emb = mx.zeros((1, 256))

                cond_tokens = conds_data.get("t3.cond_prompt_speech_tokens")

                t3_cond = T3Cond(
                    speaker_emb=speaker_emb,
                    cond_prompt_speech_tokens=cond_tokens,
                )

                gen_mlx = {}
                for k, v in conds_data.items():
                    if k.startswith("gen."):
                        gen_mlx[k.replace("gen.", "")] = v

                model._conds = Conditionals(t3_cond, gen_mlx)
                logger.info("Loaded pre-computed conditionals from safetensors")

            except Exception as e:
                logger.warning(f"Failed to load conds.safetensors: {e}")

        else:
            raise FileNotFoundError("conds.safetensors not found")
        return model

    @classmethod
    def from_local(
        cls, ckpt_dir: Union[str, Path], device: str = "cpu"
    ) -> "ChatterboxTurboTTS":
        """
        Load model from local checkpoint directory.

        Args:
            ckpt_dir: Path to checkpoint directory
            device: Device to use (ignored in MLX, always uses Metal)

        Returns:
            ChatterboxTurboTTS instance
        """
        ckpt_dir = Path(ckpt_dir)

        # Load Voice Encoder
        ve = VoiceEncoder()

        # Create T3 config for Turbo
        hp = T3Config.turbo()

        # Create T3 model
        t3 = T3(hp)

        # Create S3Gen
        s3gen = S3Gen(meanflow=True)

        # Try to load converted weights from model.safetensors
        model_weights_path = ckpt_dir / "model.safetensors"
        if model_weights_path.exists():
            logger.info(f"Loading converted weights from {model_weights_path}")
            weights = mx.load(str(model_weights_path))

            # Split weights by prefix and load into each model
            ve_weights = {
                k.replace("ve.", ""): v
                for k, v in weights.items()
                if k.startswith("ve.")
            }
            t3_weights = {
                k.replace("t3.", ""): v
                for k, v in weights.items()
                if k.startswith("t3.")
            }
            s3gen_weights = {
                k.replace("s3gen.", ""): v
                for k, v in weights.items()
                if k.startswith("s3gen.")
            }

            # Debug: Print expected vs loaded keys for VE
            from mlx.utils import tree_flatten

            ve_param_keys = [k for k, _ in tree_flatten(ve.parameters())]
            print(f"VE model expects these parameter keys: {ve_param_keys[:10]}...")
            print(f"VE weights from file: {list(ve_weights.keys())[:10]}...")

            if ve_weights:
                logger.info(f"Loading {len(ve_weights)} VE weights")
                try:
                    ve.load_weights(list(ve_weights.items()), strict=True)
                    logger.info("VE weights loaded successfully with strict=True")
                except Exception as e:
                    logger.warning(f"VE strict loading failed: {e}")
                    logger.info("Falling back to strict=False")
                    ve.load_weights(list(ve_weights.items()), strict=False)

            if t3_weights:
                logger.info(f"Loading {len(t3_weights)} T3 weights")
                try:
                    t3.load_weights(list(t3_weights.items()), strict=True)
                    logger.info("T3 weights loaded successfully with strict=True")
                except Exception as e:
                    logger.warning(f"T3 strict loading failed: {e}")
                    logger.info("Falling back to strict=False")
                    t3.load_weights(list(t3_weights.items()), strict=False)

            if s3gen_weights:
                logger.info(f"Loading {len(s3gen_weights)} S3Gen weights")
                # S3Gen has some parameters generated at init (not from weights):
                # - encoder.embed.pos_enc.pe, encoder.up_embed.pos_enc.pe (positional encodings)
                # - mel2wav.stft_window (STFT window from scipy)
                # - trim_fade (fade buffer)
                init_generated_params = {
                    "encoder.embed.pos_enc.pe",
                    "encoder.up_embed.pos_enc.pe",
                    "mel2wav.stft_window",
                    "trim_fade",
                }

                # Get all S3Gen parameter keys
                s3gen_param_keys = set(k for k, _ in tree_flatten(s3gen.parameters()))
                loadable_param_keys = s3gen_param_keys - init_generated_params

                # Find matching weights (weights that exist in model's loadable params)
                matching_weights = [
                    (k, v) for k, v in s3gen_weights.items() if k in loadable_param_keys
                ]

                # Check for any weights in file that don't match model
                unmatched_weights = set(s3gen_weights.keys()) - s3gen_param_keys
                if unmatched_weights:
                    logger.debug(
                        f"Weights in file not in model: {len(unmatched_weights)}"
                    )

                # Check for loadable params that don't have weights
                missing_weights = loadable_param_keys - set(s3gen_weights.keys())
                if missing_weights:
                    logger.warning(f"Model params without weights: {missing_weights}")

                logger.info(
                    f"Loading {len(matching_weights)} S3Gen weights (excluding {len(init_generated_params)} init-generated params)"
                )

                if matching_weights:
                    # Load with strict=False since we're intentionally excluding init-generated params
                    s3gen.load_weights(matching_weights, strict=False)

                    # Verify all expected weights were loaded
                    if len(matching_weights) == len(loadable_param_keys):
                        logger.info(
                            "S3Gen weights loaded successfully (all loadable params matched)"
                        )
                    else:
                        logger.warning(
                            f"S3Gen loaded {len(matching_weights)}/{len(loadable_param_keys)} loadable params"
                        )
                else:
                    logger.warning(
                        "No matching S3Gen weights found - model may not work correctly"
                    )

            mx.eval(ve.parameters(), t3.parameters(), s3gen.parameters())
            logger.info("Weights loaded successfully")
        else:
            logger.warning(f"No converted weights found at {model_weights_path}")
            logger.warning(
                "Run convert_weights.py first to convert PyTorch weights to MLX format"
            )

        # Load tokenizer
        try:
            from transformers import AutoTokenizer

            tokenizer = AutoTokenizer.from_pretrained(str(ckpt_dir))
            if tokenizer.pad_token is None:
                tokenizer.pad_token = tokenizer.eos_token
        except Exception as e:
            logger.warning(f"Could not load tokenizer: {e}")
            tokenizer = None

        # Load pre-computed conditionals (prefer safetensors, fallback to .pt)
        conds = None
        builtin_voice_safetensors = ckpt_dir / "conds.safetensors"
        builtin_voice_pt = ckpt_dir / "conds.pt"

        if builtin_voice_safetensors.exists():
            try:
                # Load from safetensors (pure MLX, no PyTorch dependency)
                conds_data = mx.load(str(builtin_voice_safetensors))

                # Extract t3 conditionals
                speaker_emb = conds_data.get("t3.speaker_emb")
                if speaker_emb is None:
                    speaker_emb = mx.zeros((1, 256))

                cond_tokens = conds_data.get("t3.cond_prompt_speech_tokens")

                t3_cond = T3Cond(
                    speaker_emb=speaker_emb,
                    cond_prompt_speech_tokens=cond_tokens,
                )

                # Extract gen conditionals
                gen_mlx = {}
                for k, v in conds_data.items():
                    if k.startswith("gen."):
                        gen_mlx[k.replace("gen.", "")] = v

                conds = Conditionals(t3_cond, gen_mlx)
                logger.info("Loaded pre-computed conditionals from safetensors")

            except Exception as e:
                logger.warning(f"Failed to load conds.safetensors: {e}")

        elif builtin_voice_pt.exists():
            try:
                import torch

                conds_data = torch.load(
                    builtin_voice_pt, map_location="cpu", weights_only=True
                )

                # Convert to MLX arrays
                t3_cond_dict = conds_data.get("t3", {})
                gen_dict = conds_data.get("gen", {})

                # Helper to convert PyTorch tensor to numpy (handles requires_grad)
                def to_numpy(t):
                    if hasattr(t, "detach"):
                        return t.detach().cpu().numpy()
                    elif hasattr(t, "numpy"):
                        return t.numpy()
                    return np.array(t)

                # Convert tensors to MLX arrays
                speaker_emb = t3_cond_dict.get("speaker_emb")
                if speaker_emb is not None:
                    speaker_emb = mx.array(to_numpy(speaker_emb))
                else:
                    speaker_emb = mx.array(np.zeros((1, 256), dtype=np.float32))

                cond_tokens = t3_cond_dict.get("cond_prompt_speech_tokens")
                if cond_tokens is not None:
                    cond_tokens = mx.array(to_numpy(cond_tokens).astype(np.int32))

                t3_cond = T3Cond(
                    speaker_emb=speaker_emb,
                    cond_prompt_speech_tokens=cond_tokens,
                )

                gen_mlx = {}
                for k, v in gen_dict.items():
                    if hasattr(v, "detach") or hasattr(v, "numpy"):
                        gen_mlx[k] = mx.array(to_numpy(v))
                    elif isinstance(v, (int, float)):
                        gen_mlx[k] = v

                conds = Conditionals(t3_cond, gen_mlx)
                logger.info("Loaded pre-computed conditionals from .pt file")
            except Exception as e:
                logger.warning(f"Could not load conditionals: {e}")

        return cls(t3, s3gen, ve, tokenizer, conds=conds, local_path=str(ckpt_dir))

    @classmethod
    def from_pretrained(
        cls, device: str = "cpu", weights_path: str = None
    ) -> "ChatterboxTurboTTS":
        """
        Load model from HuggingFace Hub.

        Args:
            device: Device to use (ignored in MLX)
            weights_path: Optional path to converted model.safetensors

        Returns:
            ChatterboxTurboTTS instance
        """
        try:
            from huggingface_hub import snapshot_download

            local_path = snapshot_download(
                repo_id=REPO_ID,
                token=os.getenv("HF_TOKEN") or True,
                allow_patterns=["*.safetensors", "*.json", "*.txt", "*.pt", "*.model"],
            )

            # If weights_path provided, always copy to ensure latest version is used
            if weights_path:
                import shutil

                dest = Path(local_path) / "model.safetensors"
                # Always copy to ensure we use the latest converted weights
                shutil.copy(weights_path, dest)
                logger.info(f"Copied converted weights to {dest}")

            return cls.from_local(local_path, device)

        except ImportError:
            raise ImportError(
                "Please install huggingface_hub: pip install huggingface_hub"
            )

    def norm_loudness(
        self, wav: np.ndarray, sr: int, target_lufs: float = -27
    ) -> np.ndarray:
        """Normalize audio loudness."""
        try:
            import pyloudnorm as ln

            meter = ln.Meter(sr)
            loudness = meter.integrated_loudness(wav)
            gain_db = target_lufs - loudness
            gain_linear = 10.0 ** (gain_db / 20.0)
            if math.isfinite(gain_linear) and gain_linear > 0.0:
                wav = wav * gain_linear
        except Exception as e:
            logger.warning(f"Error in norm_loudness, skipping: {e}")
        return wav

    def _extract_conditionals(
        self,
        ref_wav_24k: np.ndarray,
        ref_wav_16k: np.ndarray,
    ) -> tuple:
        """
        Extract all conditioning using pure MLX (S3Gen embeddings + T3 tokens).

        Args:
            ref_wav_24k: Reference audio at 24kHz (for S3Gen mel/decoder)
            ref_wav_16k: Reference audio at 16kHz (for S3Tokenizer)

        Returns:
            Tuple of (s3gen_ref_dict, t3_cond_prompt_tokens)
        """
        s3gen_ref_dict = {}
        t3_cond_prompt_tokens = None

        if self._s3tokenizer is not None:
            # --- S3Gen tokens (from 10s audio at 16kHz) ---
            # Trim to decoder conditioning length
            ref_16k_for_s3gen = ref_wav_16k[: int(self.DEC_COND_LEN * S3_SR / S3GEN_SR)]
            s3gen_mel = log_mel_spectrogram(mx.array(ref_16k_for_s3gen))
            s3gen_mel = mx.expand_dims(s3gen_mel, 0)  # Add batch dim
            s3gen_mel_len = mx.array([s3gen_mel.shape[2]])
            s3gen_tokens, s3gen_token_lens = self._s3tokenizer(s3gen_mel, s3gen_mel_len)

            # Get S3Gen embeddings with tokens
            s3gen_ref_dict = self.s3gen.embed_ref(
                ref_wav=mx.array(ref_wav_24k)[None, :],
                ref_sr=S3GEN_SR,
                ref_speech_tokens=s3gen_tokens,
                ref_speech_token_lens=s3gen_token_lens,
            )

            # --- T3 conditioning tokens (from encoder cond length audio) ---
            ref_16k_for_t3 = ref_wav_16k[: self.ENC_COND_LEN]
            t3_mel = log_mel_spectrogram(mx.array(ref_16k_for_t3))
            t3_mel = mx.expand_dims(t3_mel, 0)
            t3_mel_len = mx.array([t3_mel.shape[2]])
            t3_tokens, _ = self._s3tokenizer(t3_mel, t3_mel_len)

            # Limit T3 tokens to prompt length
            plen = self.t3.hp.speech_cond_prompt_len
            t3_cond_prompt_tokens = t3_tokens[:, :plen]

            logger.info("Extracted conditionals using MLX S3Tokenizer")
        else:
            logger.warning("S3Tokenizer not available - using fallback")
            # Fallback: use S3Gen's embed_ref without tokens
            s3gen_ref_dict = self.s3gen.embed_ref(
                ref_wav=mx.array(ref_wav_24k),
                ref_sr=S3GEN_SR,
            )
            # Zero tokens fallback
            plen = self.t3.hp.speech_cond_prompt_len
            if plen:
                t3_cond_prompt_tokens = mx.zeros((1, plen), dtype=mx.int32)

        return s3gen_ref_dict, t3_cond_prompt_tokens

    def prepare_conditionals(
        self,
        ref_audio: Union[str, mx.array, np.ndarray],
        sample_rate: Optional[int] = None,
        exaggeration: float = 0.5,
        norm_loudness: bool = True,
    ):
        """
        Prepare conditioning from a reference audio file or array.

        Args:
            ref_audio: Path to reference audio file or audio array (should be > 5 seconds)
            sample_rate: Sample rate of audio array (required if ref_audio is array)
            exaggeration: Emotion exaggeration factor (not used in Turbo)
            norm_loudness: Whether to normalize loudness
        """
        # Handle string path vs array input
        if isinstance(ref_audio, str):
            # Load reference audio at 24kHz for S3Gen
            ref_wav_24k, _sr = librosa.load(ref_audio, sr=S3GEN_SR)
        else:
            # Convert mx.array to numpy if needed
            if isinstance(ref_audio, mx.array):
                ref_wav_24k = np.array(ref_audio)
            else:
                ref_wav_24k = np.asarray(ref_audio)

            # Resample to S3GEN_SR if sample_rate provided and different
            input_sr = sample_rate if sample_rate is not None else S3GEN_SR
            if input_sr != S3GEN_SR:
                ref_wav_24k = librosa.resample(
                    ref_wav_24k, orig_sr=input_sr, target_sr=S3GEN_SR
                )

        assert (
            len(ref_wav_24k) / S3GEN_SR > 5.0
        ), "Audio prompt must be longer than 5 seconds!"

        if norm_loudness:
            ref_wav_24k = self.norm_loudness(ref_wav_24k, S3GEN_SR)

        # Resample to 16kHz for S3Tokenizer and voice encoder
        ref_wav_16k = librosa.resample(ref_wav_24k, orig_sr=S3GEN_SR, target_sr=S3_SR)

        # Trim 24kHz audio to decoder conditioning length
        ref_wav_24k_trimmed = ref_wav_24k[: self.DEC_COND_LEN]

        # Extract S3Gen embeddings and T3 tokens using MLX
        s3gen_ref_dict, t3_cond_prompt_tokens = self._extract_conditionals(
            ref_wav_24k_trimmed, ref_wav_16k
        )

        # Get voice encoder speaker embedding (use full 16kHz audio)
        ve_embed = self.ve.embeds_from_wavs(
            [ref_wav_16k[: self.ENC_COND_LEN]], sample_rate=S3_SR
        )
        ve_embed = mx.array(np.mean(ve_embed, axis=0, keepdims=True))

        # Create T3 conditioning
        t3_cond = T3Cond(
            speaker_emb=ve_embed,
            cond_prompt_speech_tokens=t3_cond_prompt_tokens,
            emotion_adv=(
                mx.array([[[exaggeration]]]) if self.t3.hp.emotion_adv else None
            ),
        )

        self._conds = Conditionals(t3_cond, s3gen_ref_dict)

    def generate(
        self,
        text: str,
        repetition_penalty: float = 1.2,
        min_p: float = 0.0,
        top_p: float = 0.95,
        ref_audio: Optional[Union[str, mx.array, np.ndarray]] = None,
        sample_rate: Optional[int] = None,
        exaggeration: float = 0.0,
        cfg_weight: float = 0.0,
        temperature: float = 0.8,
        top_k: int = 1000,
        norm_loudness: bool = True,
        stream: bool = False,
        streaming_interval: float = 2.0,
        split_pattern: Optional[str] = r"(?<=[.!?])\s+",
        max_tokens: int = 800,
        **kwargs,
    ) -> Generator[GenerationResult, None, None]:
        """
        Generate speech from text.

        Args:
            text: Input text to synthesize
            repetition_penalty: Penalty for repeating tokens
            min_p: Minimum probability threshold (not used in Turbo)
            top_p: Nucleus sampling threshold
            ref_audio: Optional reference audio for voice cloning (path or array)
            sample_rate: Sample rate of audio array (required if ref_audio is array)
            exaggeration: Emotion exaggeration (not used in Turbo)
            cfg_weight: Classifier-free guidance weight (not used in Turbo)
            temperature: Sampling temperature
            top_k: Top-k sampling parameter
            norm_loudness: Whether to normalize output loudness
            stream: Whether to stream audio chunks as they're generated
            streaming_interval: Time interval in seconds for streaming chunks
            split_pattern: Regex pattern to split long text into chunks (default: sentence boundaries)
            max_tokens: Maximum tokens per chunk to maintain quality (default: 900)

        Yields:
            GenerationResult with generated waveform and metrics
        """
        import re

        # If streaming is enabled, delegate to stream_generate
        if stream:
            # Convert streaming_interval (seconds) to chunk_size (tokens)
            # Each token represents ~40ms of audio (25Hz token rate)
            chunk_size = max(10, int(streaming_interval * 25))
            yield from self.stream_generate(
                text=text,
                repetition_penalty=repetition_penalty,
                min_p=min_p,
                top_p=top_p,
                ref_audio=ref_audio,
                sample_rate=sample_rate,
                exaggeration=exaggeration,
                cfg_weight=cfg_weight,
                temperature=temperature,
                top_k=top_k,
                norm_loudness=norm_loudness,
                chunk_size=chunk_size,
                split_pattern=split_pattern,
                max_tokens=max_tokens,
                **kwargs,
            )
            return

        # Prepare conditionals if audio prompt provided
        if ref_audio is not None:
            self.prepare_conditionals(
                ref_audio,
                sample_rate=sample_rate,
                exaggeration=exaggeration,
                norm_loudness=norm_loudness,
            )
        else:
            assert (
                self._conds is not None
            ), "Please `prepare_conditionals` first or specify `ref_audio`"

        # Warn about unsupported parameters
        if cfg_weight > 0.0 or exaggeration > 0.0 or min_p > 0.0:
            logger.warning(
                "CFG, min_p and exaggeration are not supported by Turbo version and will be ignored."
            )

        # Normalize text
        text = punc_norm(text)

        # Split text into chunks at sentence boundaries to keep speech tokens under limit
        # Estimate ~8 tokens per text token, so max_tokens/8 ≈ 112 text tokens
        # With ~4 chars per token, that's ~450 chars per chunk
        max_chars_per_chunk = (max_tokens // 8) * 4

        if split_pattern:
            sentences = re.split(split_pattern, text)
            chunks = []
            current_chunk = ""

            for sentence in sentences:
                sentence = sentence.strip()
                if not sentence:
                    continue

                if (
                    current_chunk
                    and len(current_chunk) + len(sentence) + 1 > max_chars_per_chunk
                ):
                    chunks.append(current_chunk.strip())
                    current_chunk = sentence
                else:
                    if current_chunk:
                        current_chunk += " " + sentence
                    else:
                        current_chunk = sentence

            if current_chunk:
                chunks.append(current_chunk.strip())
        else:
            chunks = [text]

        # Filter empty chunks
        chunks = [c for c in chunks if c.strip()]

        if not chunks:
            chunks = [text]

        start_time = time.time()
        total_token_count = 0
        total_samples = 0
        segment_idx = 0

        # Clear any accumulated cache from previous generations
        mx.clear_cache()

        for chunk in chunks:
            # Tokenize chunk
            if self.tokenizer is not None:
                text_tokens = self.tokenizer(
                    chunk, return_tensors="np", padding=True, truncation=True
                )
                text_tokens = mx.array(text_tokens.input_ids)
            else:
                logger.warning("No tokenizer available, using simple fallback")
                text_tokens = mx.array([[ord(c) for c in chunk[:512]]])

            chunk_token_count = text_tokens.shape[1]
            total_token_count += chunk_token_count

            # Generate speech tokens with T3
            speech_tokens = self.t3.inference_turbo(
                t3_cond=self._conds.t3,
                text_tokens=text_tokens,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                repetition_penalty=repetition_penalty,
                max_gen_len=max_tokens,
            )

            # Clear cache after T3 inference
            mx.clear_cache()

            # Remove OOV tokens and add silence
            speech_tokens = speech_tokens.reshape(-1)
            mask = np.where(np.array(speech_tokens) < 6561)[0].tolist()
            speech_tokens = speech_tokens[mask]
            silence = mx.array([S3GEN_SIL, S3GEN_SIL, S3GEN_SIL], dtype=mx.int32)
            speech_tokens = mx.concatenate([speech_tokens, silence])
            speech_tokens = speech_tokens[None, :]  # Add batch dimension

            # Generate waveform with S3Gen
            wav, _ = self.s3gen.inference(
                speech_tokens=speech_tokens,
                ref_dict=self._conds.gen,
                n_cfm_timesteps=2,  # Turbo uses 2 steps
            )

            # Flatten to 1D if needed
            if wav.ndim == 2:
                wav = wav.squeeze(0)

            samples = wav.shape[0]
            total_samples += samples

            # Calculate timing and metrics
            processing_time = time.time() - start_time
            audio_duration_seconds = samples / self.sample_rate

            # Format duration
            duration_hours = int(audio_duration_seconds // 3600)
            duration_mins = int((audio_duration_seconds % 3600) // 60)
            duration_secs = int(audio_duration_seconds % 60)
            duration_ms = int((audio_duration_seconds % 1) * 1000)
            duration_str = f"{duration_hours:02d}:{duration_mins:02d}:{duration_secs:02d}.{duration_ms:03d}"

            # Calculate real-time factor based on total audio generated
            total_audio_duration = total_samples / self.sample_rate
            rtf = (
                processing_time / total_audio_duration
                if total_audio_duration > 0
                else 0
            )

            yield GenerationResult(
                audio=wav,
                samples=samples,
                sample_rate=self.sample_rate,
                segment_idx=segment_idx,
                token_count=chunk_token_count,
                audio_duration=duration_str,
                real_time_factor=round(rtf, 2),
                prompt={
                    "tokens": chunk_token_count,
                    "tokens-per-sec": (
                        round(total_token_count / processing_time, 2)
                        if processing_time > 0
                        else 0
                    ),
                },
                audio_samples={
                    "samples": samples,
                    "samples-per-sec": (
                        round(total_samples / processing_time, 2)
                        if processing_time > 0
                        else 0
                    ),
                },
                processing_time_seconds=processing_time,
                peak_memory_usage=mx.get_peak_memory() / 1e9,
            )

            segment_idx += 1

            # Clear cache between chunks
            mx.clear_cache()

    def stream_generate(
        self,
        text: str,
        repetition_penalty: float = 1.2,
        min_p: float = 0.0,
        top_p: float = 0.95,
        ref_audio: Optional[Union[str, mx.array, np.ndarray]] = None,
        sample_rate: Optional[int] = None,
        exaggeration: float = 0.5,
        cfg_weight: float = 1.5,
        temperature: float = 0.8,
        top_k: int = 1000,
        norm_loudness: bool = True,
        chunk_size: int = 40,
        split_pattern: Optional[str] = r"(?<=[.!?])\s+",
        max_tokens: int = 800,
        **kwargs,
    ) -> Generator[GenerationResult, None, None]:
        """
        Stream generate speech from text, yielding audio chunks as they're generated.

        This method generates audio in a streaming fashion, yielding audio chunks
        as soon as they're ready. This allows for lower latency audio playback.

        Args:
            text: Input text to synthesize
            repetition_penalty: Penalty for repeating tokens
            min_p: Minimum probability threshold (not used in Turbo)
            top_p: Nucleus sampling threshold
            ref_audio: Optional reference audio for voice cloning (path or array)
            sample_rate: Sample rate of audio array (required if ref_audio is array)
            exaggeration: Emotion exaggeration (not used in Turbo)
            cfg_weight: Classifier-free guidance weight (not used in Turbo)
            temperature: Sampling temperature
            top_k: Top-k sampling parameter
            norm_loudness: Whether to normalize output loudness
            chunk_size: Number of tokens to accumulate before converting to audio
            split_pattern: Regex pattern to split long text into chunks (default: sentence boundaries)
            max_tokens: Maximum tokens per text chunk to maintain quality (default: 800)

        Yields:
            GenerationResult with generated audio chunks and metrics
        """
        import re

        # Prepare conditionals if audio prompt provided
        if ref_audio is not None:
            self.prepare_conditionals(
                ref_audio,
                sample_rate=sample_rate,
                exaggeration=exaggeration,
                norm_loudness=norm_loudness,
            )
        else:
            assert (
                self._conds is not None
            ), "Please `prepare_conditionals` first or specify `ref_audio`"

        # Warn about unsupported parameters
        if cfg_weight > 0.0 or exaggeration > 0.0 or min_p > 0.0:
            logger.warning(
                "CFG, min_p and exaggeration are not supported by Turbo version and will be ignored."
            )

        # Normalize text
        text = punc_norm(text)

        # Split text into chunks at sentence boundaries to keep speech tokens under limit
        max_chars_per_chunk = (max_tokens // 8) * 4

        if split_pattern:
            sentences = re.split(split_pattern, text)
            text_chunks = []
            current_chunk = ""

            for sentence in sentences:
                sentence = sentence.strip()
                if not sentence:
                    continue

                if (
                    current_chunk
                    and len(current_chunk) + len(sentence) + 1 > max_chars_per_chunk
                ):
                    text_chunks.append(current_chunk.strip())
                    current_chunk = sentence
                else:
                    if current_chunk:
                        current_chunk += " " + sentence
                    else:
                        current_chunk = sentence

            if current_chunk:
                text_chunks.append(current_chunk.strip())
        else:
            text_chunks = [text]

        # Filter empty chunks
        text_chunks = [c for c in text_chunks if c.strip()]
        if not text_chunks:
            text_chunks = [text]

        start_time = time.time()
        segment_idx = 0
        total_token_count = 0
        global_total_samples = 0

        # Clear any accumulated cache from previous generations
        mx.clear_cache()

        # Process each text chunk
        for text_chunk_idx, text_chunk in enumerate(text_chunks):
            is_last_text_chunk = text_chunk_idx == len(text_chunks) - 1

            # Tokenize chunk
            if self.tokenizer is not None:
                text_tokens = self.tokenizer(
                    text_chunk, return_tensors="np", padding=True, truncation=True
                )
                text_tokens = mx.array(text_tokens.input_ids)
            else:
                logger.warning("No tokenizer available, using simple fallback")
                text_tokens = mx.array([[ord(c) for c in text_chunk[:512]]])

            chunk_token_count = text_tokens.shape[1]
            total_token_count += chunk_token_count

            # Streaming generation - use pre-allocated buffer (in-place approach)
            max_tokens = 2000  # Pre-allocate for max expected tokens
            accumulated_tokens = mx.zeros((1, max_tokens), dtype=mx.int32)
            num_tokens = 0
            prev_audio_samples = 0
            total_samples = 0

            # Generate speech tokens in chunks
            for token_chunk, is_final in self.t3.inference_turbo_stream(
                t3_cond=self._conds.t3,
                text_tokens=text_tokens,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                repetition_penalty=repetition_penalty,
                chunk_size=chunk_size,
                max_gen_len=max_tokens,
            ):
                # In-place update of pre-allocated buffer
                chunk_len = token_chunk.shape[1]
                mx.eval(token_chunk)

                # Expand buffer if needed
                if num_tokens + chunk_len > max_tokens:
                    max_tokens = max_tokens * 2
                    new_buffer = mx.zeros((1, max_tokens), dtype=mx.int32)
                    new_buffer[:, :num_tokens] = accumulated_tokens[:, :num_tokens]
                    accumulated_tokens = new_buffer

                accumulated_tokens[:, num_tokens : num_tokens + chunk_len] = token_chunk
                num_tokens += chunk_len

                # Remove OOV tokens (convert to numpy for masking)
                token_data = np.array(accumulated_tokens[0, :num_tokens])
                mask = np.where(token_data < 6561)[0]
                valid_tokens = mx.array(token_data[mask], dtype=mx.int32)

                # Add silence tokens for final chunk of this text chunk
                if is_final:
                    silence = mx.array(
                        [S3GEN_SIL, S3GEN_SIL, S3GEN_SIL, S3GEN_SIL, S3GEN_SIL],
                        dtype=mx.int32,
                    )
                    valid_tokens = mx.concatenate([valid_tokens, silence])

                valid_tokens = valid_tokens[None, :]  # Add batch dimension

                # Convert tokens to audio
                try:
                    if is_final:
                        # For final chunk, use standard inference for complete audio
                        full_audio, _ = self.s3gen.inference(
                            speech_tokens=valid_tokens,
                            ref_dict=self._conds.gen,
                            n_cfm_timesteps=2,  # Turbo uses 2 steps
                        )
                        mx.eval(full_audio)

                        # Return only the new portion (after what was already played)
                        if full_audio.ndim == 2:
                            full_audio = full_audio.squeeze(0)

                        mx.clear_cache()

                        if prev_audio_samples > 0 and prev_audio_samples < len(
                            full_audio
                        ):
                            new_audio = mx.array(full_audio[prev_audio_samples:])
                        else:
                            new_audio = mx.array(full_audio)

                        total_samples = len(full_audio)
                    else:
                        new_audio, total_samples = self.s3gen.inference_stream(
                            speech_tokens=valid_tokens,
                            ref_dict=self._conds.gen,
                            n_cfm_timesteps=2,  # Turbo uses 2 steps
                            prev_audio_samples=prev_audio_samples,
                            is_final=is_final,
                        )
                        # Evaluate and copy to numpy to free MLX memory
                        mx.eval(new_audio)
                        mx.clear_cache()

                    # Update global sample count
                    global_total_samples += (
                        new_audio.shape[0]
                        if new_audio.ndim == 1
                        else new_audio.shape[1]
                    )

                    # Only yield if we have new audio
                    audio_len = (
                        new_audio.shape[0]
                        if new_audio.ndim == 1
                        else new_audio.shape[1]
                    )
                    if audio_len > 0:
                        # Flatten to 1D if needed
                        if new_audio.ndim == 2:
                            wav = new_audio.squeeze(0)
                        else:
                            wav = new_audio

                        # Calculate timing and metrics
                        current_time = time.time()
                        processing_time = current_time - start_time
                        samples = wav.shape[0]
                        audio_duration_seconds = samples / self.sample_rate

                        # Format duration
                        duration_hours = int(audio_duration_seconds // 3600)
                        duration_mins = int((audio_duration_seconds % 3600) // 60)
                        duration_secs = int(audio_duration_seconds % 60)
                        duration_ms = int((audio_duration_seconds % 1) * 1000)
                        duration_str = f"{duration_hours:02d}:{duration_mins:02d}:{duration_secs:02d}.{duration_ms:03d}"

                        # Calculate real-time factor
                        total_audio_duration = global_total_samples / self.sample_rate
                        rtf = (
                            processing_time / total_audio_duration
                            if total_audio_duration > 0
                            else 0
                        )

                        yield GenerationResult(
                            audio=wav,
                            samples=samples,
                            sample_rate=self.sample_rate,
                            segment_idx=segment_idx,
                            token_count=chunk_token_count,
                            audio_duration=duration_str,
                            real_time_factor=round(rtf, 2),
                            prompt={
                                "tokens": total_token_count,
                                "tokens-per-sec": (
                                    round(total_token_count / processing_time, 2)
                                    if processing_time > 0
                                    else 0
                                ),
                            },
                            audio_samples={
                                "samples": samples,
                                "samples-per-sec": (
                                    round(global_total_samples / processing_time, 2)
                                    if processing_time > 0
                                    else 0
                                ),
                            },
                            processing_time_seconds=processing_time,
                            peak_memory_usage=mx.get_peak_memory() / 1e9,
                        )

                        segment_idx += 1
                        prev_audio_samples = total_samples

                except Exception as e:
                    logger.warning(f"Error generating audio chunk: {e}")
                    continue

            # Clear cache between text chunks
            mx.clear_cache()

        # Clear cache after generation
        mx.clear_cache()



import mlx.core as mx
import mlx.nn as nn

from .config import EncoderConfig, LMConfig
from .minicpm import MiniCPMModel


class VoxCPMLocEnc(nn.Module):
    def __init__(self, config: LMConfig, input_dim: int = 64):
        super().__init__()
        self.config = config

        self.special_token = mx.random.normal((1, 1, 1, config.hidden_size))
        self.in_proj = nn.Linear(input_dim, config.hidden_size, bias=True)

        if config.vocab_size != 0:
            # warn or force?
            pass

        self.encoder = MiniCPMModel(config)

    def __call__(self, x):
        B, T, P, D = x.shape

        x = self.in_proj(x)  # (B, T, P, H)

        # Expand special token: (1, 1, 1, H) -> (B, T, 1, H)
        special_tokens = mx.broadcast_to(
            self.special_token, (B, T, 1, self.config.hidden_size)
        )

        # concat along P dimension (axis 2)
        x = mx.concatenate([special_tokens, x], axis=2)  # (B, T, P+1, H)

        # Flatten B, T -> (B*T, P+1, H)
        x = x.reshape(B * T, P + 1, -1)

        # Run encoder with bidirectional attention (is_causal=False)
        outputs, _ = self.encoder(inputs_embeds=x, is_causal=False)  # (B*T, P+1, H)

        # Take CLS token (index 0)
        cls_output = outputs[:, 0, :]  # (B*T, H)

        # Reshape back to (B, T, H)
        return cls_output.reshape(B, T, -1)



import math

import mlx.core as mx
import mlx.nn as nn


def apply_rope(
    q: mx.array,
    k: mx.array,
    offset: int = 0,
    max_period: int | float = 10000,
):
    b, t, h, d = q.shape
    if d % 2 != 0:
        raise ValueError("RoPE requires an even head dimension.")
    half = d // 2
    freqs = mx.exp(mx.arange(half, dtype=mx.float32) * (-math.log(max_period) * 2 / d))
    ts = mx.arange(t, dtype=mx.float32) + float(offset)
    ts = ts[None, :, None, None]
    q = q.reshape(b, t, h, half, 2)
    k = k.reshape(b, t, h, half, 2)

    qr = q[..., 0].astype(mx.float32)
    qi = q[..., 1].astype(mx.float32)
    kr = k[..., 0].astype(mx.float32)
    ki = k[..., 1].astype(mx.float32)

    freqs = freqs[None, None, None, :]
    rotr = mx.cos(freqs * ts)
    roti = mx.sin(freqs * ts)
    qor = qr * rotr - qi * roti
    qoi = qr * roti + qi * rotr
    kor = kr * rotr - ki * roti
    koi = kr * roti + ki * rotr

    dtype = q.dtype
    q_out = mx.stack([qor.astype(dtype), qoi.astype(dtype)], axis=-1)
    k_out = mx.stack([kor.astype(dtype), koi.astype(dtype)], axis=-1)
    return q_out.reshape(b, t, h, d), k_out.reshape(b, t, h, d)


class RotaryEmbedding(nn.Module):
    def __init__(self, max_period: int | float = 10000.0):
        super().__init__()
        self.max_period = max_period

    def __call__(self, q: mx.array, k: mx.array, offset: int):
        return apply_rope(q, k, offset, self.max_period)



# Copyright (c) 2025 Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

import math
from typing import Optional

import mlx.core as mx
import mlx.nn as nn


class RotaryEmbedding(nn.Module):
    """
    Rotary Position Embedding (RoPE) for transformer attention.

    This implementation precomputes the rotation matrices and applies them
    efficiently during the forward pass.
    """

    def __init__(
        self,
        theta: float,
        head_dim: int,
        max_seqlen: int = 1024,
        scale_factor: int = 1,
        low_freq_factor: int = 1,
        high_freq_factor: int = 32,
        old_context_len: int = 8192,
    ):
        super().__init__()
        self.theta = theta
        self.head_dim = head_dim
        self.max_seqlen = max_seqlen
        self.scale_factor = scale_factor
        self.low_freq_factor = low_freq_factor
        self.high_freq_factor = high_freq_factor
        self.old_context_len = old_context_len

        if scale_factor != 1:
            self.low_freq_wavelen = old_context_len / low_freq_factor
            self.high_freq_wavelen = old_context_len / high_freq_factor
            assert self.low_freq_wavelen >= self.high_freq_wavelen

        # Precompute rotation matrices
        self._freqs_cis = None
        self.reset_parameters()

    def reset_parameters(self):
        """Precompute rotation matrices."""
        self._freqs_cis = self._precompute_freqs_cis(
            dim=self.head_dim,
            end=self.max_seqlen,
            theta=self.theta,
        )

    def _apply_scaling(self, freqs: mx.array) -> mx.array:
        """Apply frequency scaling for context extension."""
        if self.scale_factor == 1:
            return freqs

        new_freqs = []
        for i in range(freqs.shape[0]):
            freq = freqs[i].item()
            wavelen = 2 * math.pi / freq
            if wavelen < self.high_freq_wavelen:
                new_freqs.append(freq)
            elif wavelen > self.low_freq_wavelen:
                new_freqs.append(freq / self.scale_factor)
            else:
                smooth = (self.old_context_len / wavelen - self.low_freq_factor) / (
                    self.high_freq_factor - self.low_freq_factor
                )
                new_freqs.append(
                    (1 - smooth) * freq / self.scale_factor + smooth * freq
                )
        return mx.array(new_freqs, dtype=freqs.dtype)

    def _precompute_freqs_cis(
        self,
        dim: int,
        end: int,
        theta: float = 10000.0,
    ) -> mx.array:
        """
        Precompute frequency tensor for complex exponentials (cis) with rotation matrices.

        Uses the same 2x2 rotation matrix approach as PyTorch SAM-Audio.

        Args:
            dim: Dimension of the head
            end: Maximum sequence length
            theta: Base for the frequency computation

        Returns:
            Precomputed frequency tensor with shape (1, end, 1, dim//2, 2, 2)
        """
        freqs = 1.0 / (
            theta ** (mx.arange(0, dim, 2, dtype=mx.float32)[: (dim // 2)] / dim)
        )
        freqs = self._apply_scaling(freqs)

        t = mx.arange(end, dtype=mx.float32)
        freqs = mx.outer(t, freqs)  # (end, dim//2)

        cos = mx.cos(freqs)  # (end, dim//2)
        sin = mx.sin(freqs)  # (end, dim//2)

        # Create 2x2 rotation matrix: [[cos, -sin], [sin, cos]]
        # Stack as (cos, -sin, sin, cos) and reshape to (..., 2, 2)
        freqs_cis = mx.stack([cos, -sin, sin, cos], axis=-1)
        freqs_cis = freqs_cis.reshape(*freqs.shape, 2, 2)  # (end, dim//2, 2, 2)

        # Add batch and head dims: (end, dim//2, 2, 2) -> (1, end, 1, dim//2, 2, 2)
        freqs_cis = mx.expand_dims(mx.expand_dims(freqs_cis, 0), 2)

        return freqs_cis

    def __call__(
        self,
        x: mx.array,
        bhle: bool = False,
        offset: int = 0,
    ) -> mx.array:
        """
        Apply rotary embeddings to input tensor.

        Uses 2x2 rotation matrix multiplication matching PyTorch SAM-Audio.

        Args:
            x: Input tensor of shape (B, L, H, E) or (B, H, L, E) if bhle=True
            bhle: If True, input is in (B, H, L, E) format
            offset: Position offset for incremental decoding

        Returns:
            Rotary-embedded tensor of same shape as input
        """
        if bhle:
            # (B, H, L, E) -> (B, L, H, E)
            x = mx.transpose(x, (0, 2, 1, 3))

        batch_size, seq_len, num_heads, head_dim = x.shape

        # Reshape to adjacent pairs: (B, L, H, E) -> (B, L, H, E/2, 1, 2)
        x_ = x.reshape(batch_size, seq_len, num_heads, head_dim // 2, 1, 2)

        # Get rotation matrix for sequence length
        # _freqs_cis shape: (1, max_seqlen, 1, dim//2, 2, 2)
        freqs_cis = self._freqs_cis[:, offset : offset + seq_len]

        # Apply rotation matrix multiplication
        # x_ shape: (B, L, H, E/2, 1, 2)
        # freqs_cis shape: (1, L, 1, E/2, 2, 2)
        # Result: (B, L, H, E/2, 1, 2) * (1, L, 1, E/2, 2, 2) -> sum over last dim -> (B, L, H, E/2, 2)
        x_out = (x_ * freqs_cis).sum(axis=-1)  # (B, L, H, E/2, 2)
        x_out = x_out.reshape(batch_size, seq_len, num_heads, head_dim)

        if bhle:
            # (B, L, H, E) -> (B, H, L, E)
            x_out = mx.transpose(x_out, (0, 2, 1, 3))

        return x_out.astype(x.dtype)


def apply_rotary_emb(
    xq: mx.array,
    xk: mx.array,
    freqs_cos: mx.array,
    freqs_sin: mx.array,
) -> tuple:
    """
    Apply rotary embeddings to query and key tensors.

    Args:
        xq: Query tensor (B, L, H, D)
        xk: Key tensor (B, L, H, D)
        freqs_cos: Cosine frequencies (L, D//2)
        freqs_sin: Sine frequencies (L, D//2)

    Returns:
        Tuple of rotated (query, key) tensors
    """
    # Reshape frequencies for broadcasting
    cos = mx.expand_dims(mx.expand_dims(freqs_cos, 0), 2)  # (1, L, 1, D//2)
    sin = mx.expand_dims(mx.expand_dims(freqs_sin, 0), 2)  # (1, L, 1, D//2)

    head_dim = xq.shape[-1]

    # Split into halves
    xq1, xq2 = xq[..., : head_dim // 2], xq[..., head_dim // 2 :]
    xk1, xk2 = xk[..., : head_dim // 2], xk[..., head_dim // 2 :]

    # Apply rotation
    xq_out = mx.concatenate([xq1 * cos - xq2 * sin, xq1 * sin + xq2 * cos], axis=-1)
    xk_out = mx.concatenate([xk1 * cos - xk2 * sin, xk1 * sin + xk2 * cos], axis=-1)

    return xq_out.astype(xq.dtype), xk_out.astype(xk.dtype)



from typing import Dict, Optional

import mlx.core as mx
import mlx.nn as nn


class CausalMaskedDiffWithXvec(nn.Module):

    def __init__(
        self,
        input_size: int = 512,
        output_size: int = 80,
        spk_embed_dim: int = 192,
        output_type: str = "mel",
        vocab_size: int = 6561,
        input_frame_rate: int = 25,
        only_mask_loss: bool = True,
        token_mel_ratio: int = 2,
        pre_lookahead_len: int = 3,
        n_timesteps: int = 10,
        encoder: nn.Module = None,
        decoder: nn.Module = None,
        decoder_conf: Dict = None,
        mel_feat_conf: Dict = None,
    ):

        super().__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.decoder_conf = decoder_conf or {}
        self.mel_feat_conf = mel_feat_conf or {}
        self.vocab_size = vocab_size
        self.output_type = output_type
        self.input_frame_rate = input_frame_rate
        self.n_timesteps = n_timesteps

        self.input_embedding = nn.Embedding(vocab_size, input_size)
        self.spk_embed_affine_layer = nn.Linear(spk_embed_dim, output_size)
        self.encoder = encoder
        self.encoder_proj = nn.Linear(self.encoder.output_size(), output_size)
        self.decoder = decoder
        self.only_mask_loss = only_mask_loss
        self.token_mel_ratio = token_mel_ratio
        self.pre_lookahead_len = pre_lookahead_len

    def inference(
        self,
        token: mx.array,
        token_len: mx.array,
        prompt_token: mx.array,
        prompt_token_len: mx.array,
        prompt_feat: mx.array,
        prompt_feat_len: mx.array,
        embedding: mx.array,
        finalize: bool,
        n_timesteps: Optional[int] = None,
        streaming: bool = False,
    ):

        assert token.shape[0] == 1

        # Speaker embedding projection
        norm = mx.linalg.norm(embedding, axis=1, keepdims=True)
        embedding = embedding / (norm + 1e-8)  # Normalize with epsilon
        embedding = self.spk_embed_affine_layer(embedding)

        # Concatenate prompt and new tokens
        token = mx.concatenate([prompt_token, token], axis=1)
        token_len = prompt_token_len + token_len

        # Create mask
        batch_size = token_len.shape[0]
        max_len = int(mx.max(token_len).item())
        seq_range = mx.arange(max_len)
        seq_range_expand = mx.expand_dims(seq_range, 0)
        seq_range_expand = mx.broadcast_to(seq_range_expand, (batch_size, max_len))
        seq_length_expand = mx.expand_dims(token_len, -1)
        mask = seq_range_expand < seq_length_expand
        mask = mx.expand_dims(mask, -1).astype(embedding.dtype)

        # Embed tokens
        num_embeddings = self.input_embedding.weight.shape[0]
        token = mx.clip(token, 0, num_embeddings - 1)
        token = self.input_embedding(token) * mask

        h, h_lengths = self.encoder(token, token_len, streaming=streaming)

        if not finalize:
            h = h[:, : -self.pre_lookahead_len * self.token_mel_ratio]

        mel_len1 = prompt_feat.shape[1]
        mel_len2 = h.shape[1] - prompt_feat.shape[1]
        h = self.encoder_proj(h)

        # Get conditions (prompt mel features)
        conds = mx.zeros([1, mel_len1 + mel_len2, self.output_size], dtype=h.dtype)
        conds[:, :mel_len1] = prompt_feat
        conds = mx.transpose(conds, [0, 2, 1])  # (1, D, T)

        # Create mask for decoder (float dtype for multiplication)
        total_len = mel_len1 + mel_len2
        mask = mx.ones([1, 1, total_len], dtype=h.dtype)

        # Generate mel features with streaming parameter
        feat, _ = self.decoder(
            mu=mx.transpose(h, [0, 2, 1]),
            mask=mask,
            spks=embedding,
            cond=conds,
            n_timesteps=n_timesteps if n_timesteps is not None else self.n_timesteps,
            streaming=streaming,
        )

        feat = feat[:, :, mel_len1:]
        assert feat.shape[2] == mel_len2

        return feat, None



# Copyright 2025, Prince Canuma and contributors (https://github.com/Blaizzy/mlx-audio)

import math
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from tqdm import tqdm

from mlx_audio.stt.models.base import STTOutput

from .config import AudioEncoderConfig, ModelConfig, TextConfig


@dataclass
class StreamingResult:
    """Result object for streaming transcription.

    Attributes:
        text: Decoded text for this emission.
        is_final: True if this is a final (committed) result, False if partial.
        start_time: Start timestamp in seconds.
        end_time: End timestamp in seconds.
        language: Language of the transcription.
    """

    text: str
    is_final: bool
    start_time: float
    end_time: float
    language: str = "en"


def split_audio_into_chunks(
    wav: np.ndarray,
    sr: int,
    chunk_duration: float = 1200.0,
    min_chunk_duration: float = 1.0,
    search_expand_sec: float = 5.0,
    min_window_ms: float = 100.0,
) -> List[Tuple[np.ndarray, float]]:
    """Split long audio into chunks at low-energy boundaries.

    Args:
        wav: Audio waveform (1D numpy array).
        sr: Sample rate.
        chunk_duration: Maximum chunk duration in seconds (default: 1200 = 20 min).
        min_chunk_duration: Minimum chunk duration in seconds (default: 1.0).
        search_expand_sec: Window to search for silence around cut point.
        min_window_ms: Minimum window size for energy calculation.

    Returns:
        List of (chunk_waveform, offset_seconds) tuples.
    """
    # Ensure mono
    if wav.ndim > 1:
        wav = wav.mean(axis=-1) if wav.shape[-1] <= 2 else wav.mean(axis=0)

    total_samples = len(wav)
    total_sec = total_samples / sr

    # If short enough, return as-is
    if total_sec <= chunk_duration:
        # Pad if too short
        if total_sec < min_chunk_duration:
            min_samples = int(min_chunk_duration * sr)
            wav = np.pad(wav, (0, min_samples - len(wav)))
        return [(wav, 0.0)]

    chunks = []
    start_sample = 0
    max_chunk_samples = int(chunk_duration * sr)
    search_samples = int(search_expand_sec * sr)
    min_window_samples = int(min_window_ms * sr / 1000)

    while start_sample < total_samples:
        end_sample = min(start_sample + max_chunk_samples, total_samples)

        # If this is the last chunk, take the rest
        if end_sample >= total_samples:
            chunk = wav[start_sample:total_samples]
            offset_sec = start_sample / sr
            # Pad if too short
            if len(chunk) < min_chunk_duration * sr:
                min_samples = int(min_chunk_duration * sr)
                chunk = np.pad(chunk, (0, min_samples - len(chunk)))
            chunks.append((chunk, offset_sec))
            break

        # Search for low-energy point around the cut
        search_start = max(start_sample, end_sample - search_samples)
        search_end = min(total_samples, end_sample + search_samples)
        search_region = wav[search_start:search_end]

        # Calculate energy using sliding window
        if len(search_region) > min_window_samples:
            energy = np.convolve(
                search_region**2,
                np.ones(min_window_samples) / min_window_samples,
                mode="valid",
            )
            # Find minimum energy point
            min_idx = np.argmin(energy) + min_window_samples // 2
            cut_sample = search_start + min_idx
        else:
            cut_sample = end_sample

        # Ensure we make progress
        cut_sample = max(cut_sample, start_sample + sr)  # At least 1 second

        chunk = wav[start_sample:cut_sample]
        offset_sec = start_sample / sr

        # Pad if too short
        if len(chunk) < min_chunk_duration * sr:
            min_samples = int(min_chunk_duration * sr)
            chunk = np.pad(chunk, (0, min_samples - len(chunk)))

        chunks.append((chunk, offset_sec))
        start_sample = cut_sample

    return chunks


def create_additive_causal_mask(N: int, offset: int = 0) -> mx.array:
    """Create an additive causal attention mask."""
    rinds = mx.arange(offset + N)
    linds = mx.arange(offset, offset + N) if offset else rinds
    mask = linds[:, None] < rinds[None]
    return mask * -1e9


def _floor_div(a: mx.array, b: int) -> mx.array:
    """Floor division matching Python semantics."""
    return mx.floor(a.astype(mx.float32) / b).astype(mx.int32)


def _get_feat_extract_output_lengths(input_lengths: mx.array) -> mx.array:
    """Compute output length of the convolutional layers."""
    input_lengths_leave = input_lengths % 100
    feat_lengths = _floor_div(input_lengths_leave - 1, 2) + 1
    output_lengths = (
        _floor_div(_floor_div(feat_lengths - 1, 2) + 1 - 1, 2)
        + 1
        + (input_lengths // 100) * 13
    )
    return output_lengths


class SinusoidalPositionEmbedding(nn.Module):
    """Sinusoidal position embeddings for the audio encoder."""

    def __init__(self, length: int, channels: int, max_timescale: float = 10000.0):
        super().__init__()
        if channels % 2 != 0:
            raise ValueError("SinusoidalPositionEmbedding needs even channels input")

        log_timescale_increment = math.log(max_timescale) / (channels // 2 - 1)
        inv_timescales = mx.exp(
            -log_timescale_increment * mx.arange(channels // 2, dtype=mx.float32)
        )
        positions = mx.arange(length, dtype=mx.float32)[:, None]
        scaled_time = positions * inv_timescales[None, :]
        self._positional_embedding = mx.concatenate(
            [mx.sin(scaled_time), mx.cos(scaled_time)], axis=1
        )

    def __call__(self, seqlen: int) -> mx.array:
        return self._positional_embedding[:seqlen, :]


class AudioAttention(nn.Module):
    """Multi-headed attention for audio encoder."""

    def __init__(self, config: AudioEncoderConfig):
        super().__init__()
        self.embed_dim = config.d_model
        self.num_heads = config.encoder_attention_heads
        self.head_dim = self.embed_dim // self.num_heads
        self.scaling = self.head_dim**-0.5

        if (self.head_dim * self.num_heads) != self.embed_dim:
            raise ValueError(
                f"embed_dim must be divisible by num_heads (got `embed_dim`: {self.embed_dim}"
                f" and `num_heads`: {self.num_heads})."
            )

        self.q_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)
        self.k_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)
        self.v_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)
        self.out_proj = nn.Linear(self.embed_dim, self.embed_dim, bias=True)

    def __call__(
        self,
        hidden_states: mx.array,
        mask: Optional[mx.array] = None,
    ) -> mx.array:
        bsz, seq_len, _ = hidden_states.shape

        query_states = self.q_proj(hidden_states) * self.scaling
        key_states = self.k_proj(hidden_states)
        value_states = self.v_proj(hidden_states)

        query_states = query_states.reshape(
            bsz, seq_len, self.num_heads, self.head_dim
        ).transpose(0, 2, 1, 3)
        key_states = key_states.reshape(
            bsz, seq_len, self.num_heads, self.head_dim
        ).transpose(0, 2, 1, 3)
        value_states = value_states.reshape(
            bsz, seq_len, self.num_heads, self.head_dim
        ).transpose(0, 2, 1, 3)

        attn_output = mx.fast.scaled_dot_product_attention(
            query_states, key_states, value_states, scale=1.0, mask=mask
        )

        attn_output = attn_output.transpose(0, 2, 1, 3).reshape(
            bsz, seq_len, self.embed_dim
        )
        return self.out_proj(attn_output)


class AudioEncoderLayer(nn.Module):
    """A single transformer encoder layer for audio."""

    def __init__(self, config: AudioEncoderConfig):
        super().__init__()
        self.embed_dim = config.d_model
        self.self_attn = AudioAttention(config)
        self.self_attn_layer_norm = nn.LayerNorm(self.embed_dim)
        self.fc1 = nn.Linear(self.embed_dim, config.encoder_ffn_dim)
        self.fc2 = nn.Linear(config.encoder_ffn_dim, self.embed_dim)
        self.final_layer_norm = nn.LayerNorm(self.embed_dim)

    def __call__(
        self,
        hidden_states: mx.array,
        mask: Optional[mx.array] = None,
    ) -> mx.array:
        residual = hidden_states
        hidden_states = self.self_attn_layer_norm(hidden_states)
        hidden_states = self.self_attn(hidden_states, mask=mask)
        hidden_states = residual + hidden_states

        residual = hidden_states
        hidden_states = self.final_layer_norm(hidden_states)
        hidden_states = nn.gelu(self.fc1(hidden_states))
        hidden_states = self.fc2(hidden_states)
        hidden_states = residual + hidden_states

        return hidden_states


class AudioEncoder(nn.Module):
    """Qwen3-ASR Audio Encoder with Conv2d frontend and transformer layers."""

    def __init__(self, config: AudioEncoderConfig):
        super().__init__()
        self.config = config
        embed_dim = config.d_model
        self.num_mel_bins = config.num_mel_bins
        self.max_source_positions = config.max_source_positions
        self.embed_scale = math.sqrt(embed_dim) if config.scale_embedding else 1.0
        self.n_window = config.n_window
        self.n_window_infer = config.n_window_infer
        self.conv_chunksize = config.conv_chunksize

        self.conv2d1 = nn.Conv2d(
            1, config.downsample_hidden_size, kernel_size=3, stride=2, padding=1
        )
        self.conv2d2 = nn.Conv2d(
            config.downsample_hidden_size,
            config.downsample_hidden_size,
            kernel_size=3,
            stride=2,
            padding=1,
        )
        self.conv2d3 = nn.Conv2d(
            config.downsample_hidden_size,
            config.downsample_hidden_size,
            kernel_size=3,
            stride=2,
            padding=1,
        )

        freq_after_conv = ((((config.num_mel_bins + 1) // 2) + 1) // 2 + 1) // 2
        self.conv_out = nn.Linear(
            config.downsample_hidden_size * freq_after_conv, embed_dim, bias=False
        )
        self.positional_embedding = SinusoidalPositionEmbedding(
            self.max_source_positions, embed_dim
        )
        self.layers = [AudioEncoderLayer(config) for _ in range(config.encoder_layers)]
        self.ln_post = nn.LayerNorm(embed_dim)
        self.proj1 = nn.Linear(embed_dim, embed_dim)
        self.proj2 = nn.Linear(embed_dim, config.output_dim)

    def _create_block_attention_mask(
        self, seq_len: int, cu_seqlens: List[int], dtype: mx.Dtype
    ) -> mx.array:
        """Create attention mask for ragged/block attention."""
        mask = mx.full((seq_len, seq_len), -1e9, dtype=dtype)
        for i in range(len(cu_seqlens) - 1):
            start = cu_seqlens[i]
            end = cu_seqlens[i + 1]
            mask[start:end, start:end] = 0.0
        return mask

    def __call__(
        self,
        input_features: mx.array,
        feature_attention_mask: Optional[mx.array] = None,
    ) -> mx.array:
        if feature_attention_mask is not None:
            feature_lens = feature_attention_mask.sum(axis=-1).astype(mx.int32)
        else:
            feature_lens = mx.array(
                [input_features.shape[-1]] * input_features.shape[0], dtype=mx.int32
            )

        feature_lens_np = np.array(feature_lens)
        aftercnn_lens = _get_feat_extract_output_lengths(feature_lens)
        chunk_size = self.n_window * 2
        chunk_num = np.ceil(feature_lens_np / chunk_size).astype(np.int32)

        chunk_lengths = []
        for i in range(len(feature_lens_np)):
            num_chunks = int(chunk_num[i])
            feat_len = int(feature_lens_np[i])
            for j in range(num_chunks):
                if j == num_chunks - 1:
                    remainder = feat_len % chunk_size
                    chunk_lengths.append(chunk_size if remainder == 0 else remainder)
                else:
                    chunk_lengths.append(chunk_size)

        chunk_lengths = np.array(chunk_lengths, dtype=np.int32)

        chunks = []
        for i in range(len(feature_lens_np)):
            feat = input_features[i]
            feat_len = int(feature_lens_np[i])
            num_chunks = int(chunk_num[i])
            pos = 0
            for j in range(num_chunks):
                if j == num_chunks - 1:
                    remainder = feat_len % chunk_size
                    clen = chunk_size if remainder == 0 else remainder
                else:
                    clen = chunk_size
                chunk = feat[:, pos : pos + clen]
                chunks.append(chunk)
                pos += clen

        max_chunk_len = int(max(chunk_lengths))
        padded_chunks = []
        for i, chunk in enumerate(chunks):
            clen = int(chunk_lengths[i])
            if clen < max_chunk_len:
                pad_width = max_chunk_len - clen
                chunk = mx.pad(chunk, [(0, 0), (0, pad_width)])
            padded_chunks.append(chunk)

        padded_feature = mx.stack(padded_chunks, axis=0)

        feature_lens_after_cnn = _get_feat_extract_output_lengths(
            mx.array(chunk_lengths)
        )
        feature_lens_after_cnn_np = np.array(feature_lens_after_cnn)
        max_len_after_cnn = int(feature_lens_after_cnn_np.max())

        padded_mask_after_cnn = np.zeros(
            (len(chunk_lengths), max_len_after_cnn), dtype=bool
        )
        for i, length in enumerate(feature_lens_after_cnn_np):
            padded_mask_after_cnn[i, : int(length)] = True

        x = padded_feature[:, :, :, None]
        x = nn.gelu(self.conv2d1(x))
        x = nn.gelu(self.conv2d2(x))
        x = nn.gelu(self.conv2d3(x))

        b, f, t, c = x.shape
        x = x.transpose(0, 2, 3, 1).reshape(b, t, c * f)
        x = self.conv_out(x)

        pos_emb = self.positional_embedding(x.shape[1])
        x = x + pos_emb[None, :, :]

        hidden_list = []
        for i in range(x.shape[0]):
            valid_len = int(feature_lens_after_cnn_np[i])
            hidden_list.append(x[i, :valid_len])

        hidden_states = mx.concatenate(hidden_list, axis=0)

        aftercnn_lens_np = np.array(aftercnn_lens)
        window_aftercnn = max_len_after_cnn * (
            self.n_window_infer // (self.n_window * 2)
        )

        cu_chunk_lens = [0]
        for cnn_len in aftercnn_lens_np:
            cnn_len = int(cnn_len)
            num_full_windows = cnn_len // window_aftercnn
            for _ in range(num_full_windows):
                cu_chunk_lens.append(window_aftercnn)
            remainder = cnn_len % window_aftercnn
            if remainder != 0:
                cu_chunk_lens.append(remainder)

        cu_seqlens = np.cumsum(cu_chunk_lens).tolist()

        seq_len = hidden_states.shape[0]
        attention_mask = self._create_block_attention_mask(
            seq_len, cu_seqlens, hidden_states.dtype
        )
        attention_mask = attention_mask[None, None, :, :]

        hidden_states = hidden_states[None, :, :]

        for layer in self.layers:
            hidden_states = layer(hidden_states, mask=attention_mask)

        hidden_states = hidden_states[0]
        hidden_states = self.ln_post(hidden_states)
        hidden_states = nn.gelu(self.proj1(hidden_states))
        hidden_states = self.proj2(hidden_states)

        return hidden_states


class TextAttention(nn.Module):
    """Multi-headed attention for text decoder with Q/K norms."""

    def __init__(self, config: TextConfig, layer_idx: int):
        super().__init__()
        self.config = config
        self.layer_idx = layer_idx
        self.hidden_size = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.num_kv_heads = config.num_key_value_heads
        self.head_dim = config.head_dim
        self.scale = self.head_dim**-0.5

        self.q_proj = nn.Linear(
            config.hidden_size, self.num_heads * self.head_dim, bias=False
        )
        self.k_proj = nn.Linear(
            config.hidden_size, self.num_kv_heads * self.head_dim, bias=False
        )
        self.v_proj = nn.Linear(
            config.hidden_size, self.num_kv_heads * self.head_dim, bias=False
        )
        self.o_proj = nn.Linear(
            self.num_heads * self.head_dim, config.hidden_size, bias=False
        )

        self.q_norm = nn.RMSNorm(self.head_dim, eps=config.rms_norm_eps)
        self.k_norm = nn.RMSNorm(self.head_dim, eps=config.rms_norm_eps)
        self.rope = nn.RoPE(self.head_dim, traditional=False, base=config.rope_theta)

    def __call__(
        self,
        hidden_states: mx.array,
        cache: Optional[Any] = None,
    ) -> mx.array:
        B, L, _ = hidden_states.shape

        queries = self.q_proj(hidden_states)
        keys = self.k_proj(hidden_states)
        values = self.v_proj(hidden_states)

        queries = queries.reshape(B, L, self.num_heads, self.head_dim)
        keys = keys.reshape(B, L, self.num_kv_heads, self.head_dim)
        values = values.reshape(B, L, self.num_kv_heads, self.head_dim)

        queries = self.q_norm(queries)
        keys = self.k_norm(keys)

        queries = queries.transpose(0, 2, 1, 3)
        keys = keys.transpose(0, 2, 1, 3)
        values = values.transpose(0, 2, 1, 3)

        if cache is not None:
            offset = cache.offset
            queries = self.rope(queries, offset=offset)
            keys = self.rope(keys, offset=offset)
        else:
            offset = 0
            queries = self.rope(queries)
            keys = self.rope(keys)

        if cache is not None:
            keys, values = cache.update_and_fetch(keys, values)

        query_len = queries.shape[2]
        mask = create_additive_causal_mask(query_len, offset=offset).astype(
            queries.dtype
        )

        output = mx.fast.scaled_dot_product_attention(
            queries, keys, values, scale=self.scale, mask=mask
        )

        output = output.transpose(0, 2, 1, 3).reshape(B, query_len, -1)
        return self.o_proj(output)


class TextMLP(nn.Module):
    """MLP for text decoder with SwiGLU activation."""

    def __init__(self, config: TextConfig):
        super().__init__()
        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size
        self.gate_proj = nn.Linear(
            config.hidden_size, config.intermediate_size, bias=False
        )
        self.up_proj = nn.Linear(
            config.hidden_size, config.intermediate_size, bias=False
        )
        self.down_proj = nn.Linear(
            config.intermediate_size, config.hidden_size, bias=False
        )

    def __call__(self, x: mx.array) -> mx.array:
        return self.down_proj(nn.silu(self.gate_proj(x)) * self.up_proj(x))


class TextDecoderLayer(nn.Module):
    """A single transformer decoder layer."""

    def __init__(self, config: TextConfig, layer_idx: int):
        super().__init__()
        self.hidden_size = config.hidden_size
        self.self_attn = TextAttention(config, layer_idx)
        self.mlp = TextMLP(config)
        self.input_layernorm = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.post_attention_layernorm = nn.RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )

    def __call__(
        self,
        hidden_states: mx.array,
        cache: Optional[Any] = None,
    ) -> mx.array:
        residual = hidden_states
        hidden_states = self.input_layernorm(hidden_states)
        hidden_states = self.self_attn(hidden_states, cache=cache)
        hidden_states = residual + hidden_states

        residual = hidden_states
        hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)
        hidden_states = residual + hidden_states

        return hidden_states


class TextModel(nn.Module):
    """Text decoder model (Qwen3-based)."""

    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.vocab_size = config.vocab_size
        self.num_hidden_layers = config.num_hidden_layers
        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size)
        self.layers = [
            TextDecoderLayer(config, i) for i in range(config.num_hidden_layers)
        ]
        self.norm = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def __call__(
        self,
        input_ids: Optional[mx.array] = None,
        inputs_embeds: Optional[mx.array] = None,
        cache: Optional[List[Any]] = None,
    ) -> mx.array:
        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        hidden_states = inputs_embeds

        if cache is None:
            cache = [None] * len(self.layers)

        for i, layer in enumerate(self.layers):
            hidden_states = layer(hidden_states, cache=cache[i])

        return self.norm(hidden_states)


class Qwen3ASRModel(nn.Module):
    """Qwen3-ASR Model for speech recognition."""

    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vocab_size = config.text_config.vocab_size
        self.audio_tower = AudioEncoder(config.audio_config)
        self.model = TextModel(config.text_config)

        if config.text_config.tie_word_embeddings:
            self.lm_head = None
        else:
            self.lm_head = nn.Linear(
                config.text_config.hidden_size,
                config.text_config.vocab_size,
                bias=False,
            )

    def get_audio_features(
        self,
        input_features: mx.array,
        feature_attention_mask: Optional[mx.array] = None,
    ) -> mx.array:
        """Encode audio features."""
        return self.audio_tower(input_features, feature_attention_mask)

    def _build_inputs_embeds(
        self,
        input_ids: mx.array,
        audio_features: mx.array,
    ) -> mx.array:
        """Build input embeddings with audio features merged in."""
        inputs_embeds = self.model.embed_tokens(input_ids)
        audio_features = audio_features.astype(inputs_embeds.dtype)

        audio_token_mask = input_ids == self.config.audio_token_id

        if audio_token_mask.any():
            batch_size, seq_len, hidden_dim = inputs_embeds.shape
            flat_mask = audio_token_mask.flatten()
            flat_mask_np = np.array(flat_mask)
            audio_indices = np.where(flat_mask_np)[0]

            if len(audio_indices) > 0 and audio_features.shape[0] > 0:
                num_to_replace = min(len(audio_indices), audio_features.shape[0])
                flat_embeds = inputs_embeds.reshape(-1, hidden_dim)

                result_list = []
                audio_idx = 0
                for i in range(flat_embeds.shape[0]):
                    if audio_idx < num_to_replace and i == audio_indices[audio_idx]:
                        result_list.append(audio_features[audio_idx])
                        audio_idx += 1
                    else:
                        result_list.append(flat_embeds[i])

                inputs_embeds = mx.stack(result_list, axis=0).reshape(
                    batch_size, seq_len, hidden_dim
                )

        return inputs_embeds

    def _forward_with_embeds(
        self,
        inputs_embeds: mx.array,
        cache: Optional[List[Any]] = None,
    ) -> mx.array:
        """Forward pass using pre-computed embeddings."""
        hidden_states = self.model(inputs_embeds=inputs_embeds, cache=cache)

        if self.lm_head is not None:
            logits = self.lm_head(hidden_states)
        else:
            logits = self.model.embed_tokens.as_linear(hidden_states)

        return logits

    def __call__(
        self,
        input_ids: mx.array,
        input_embeddings: Optional[mx.array] = None,
        input_features: Optional[mx.array] = None,
        feature_attention_mask: Optional[mx.array] = None,
        cache: Optional[List[Any]] = None,
    ) -> mx.array:
        if input_embeddings is None:
            inputs_embeds = self.model.embed_tokens(input_ids)
        else:
            inputs_embeds = input_embeddings

        if input_features is not None and (
            cache is None or cache[0] is None or cache[0].offset == 0
        ):
            audio_features = self.get_audio_features(
                input_features, feature_attention_mask
            )
            audio_features = audio_features.astype(inputs_embeds.dtype)

            audio_token_mask = input_ids == self.config.audio_token_id

            if audio_token_mask.any():
                batch_size, seq_len, hidden_dim = inputs_embeds.shape

                flat_mask = audio_token_mask.flatten()
                flat_mask_np = np.array(flat_mask)
                audio_indices = np.where(flat_mask_np)[0]

                if len(audio_indices) > 0 and audio_features.shape[0] > 0:
                    num_to_replace = min(len(audio_indices), audio_features.shape[0])
                    flat_embeds = inputs_embeds.reshape(-1, hidden_dim)

                    result_list = []
                    audio_idx = 0
                    for i in range(flat_embeds.shape[0]):
                        if audio_idx < num_to_replace and i == audio_indices[audio_idx]:
                            result_list.append(audio_features[audio_idx])
                            audio_idx += 1
                        else:
                            result_list.append(flat_embeds[i])

                    inputs_embeds = mx.stack(result_list, axis=0).reshape(
                        batch_size, seq_len, hidden_dim
                    )

        hidden_states = self.model(inputs_embeds=inputs_embeds, cache=cache)

        if self.lm_head is not None:
            logits = self.lm_head(hidden_states)
        else:
            logits = self.model.embed_tokens.as_linear(hidden_states)

        return logits

    @property
    def layers(self):
        return self.model.layers

    @property
    def sample_rate(self) -> int:
        """Sample rate for audio input."""
        return 16000

    def make_cache(self) -> List[Any]:
        """Create KV cache for generation."""
        from mlx_lm.models.cache import KVCache

        return [KVCache() for _ in range(self.config.text_config.num_hidden_layers)]

    @staticmethod
    def sanitize(weights: Dict[str, mx.array]) -> Dict[str, mx.array]:
        """Sanitize weights from HuggingFace/PyTorch format to MLX format."""
        sanitized = {}
        is_formatted = not any(k.startswith("thinker.") for k in weights.keys())

        for k, v in weights.items():
            if k.startswith("thinker."):
                k = k[len("thinker.") :]

            if k == "lm_head.weight":
                continue

            if (
                not is_formatted
                and "conv2d" in k
                and "weight" in k
                and len(v.shape) == 4
            ):
                v = v.transpose(0, 2, 3, 1)

            sanitized[k] = v

        return sanitized

    def model_quant_predicate(self, p: str, m: nn.Module) -> bool:
        """Determine which layers to quantize."""
        return not p.startswith("audio_tower")

    @classmethod
    def post_load_hook(
        cls, model: "Qwen3ASRModel", model_path: Path
    ) -> "Qwen3ASRModel":
        """Hook called after model weights are loaded."""
        import transformers
        from transformers import AutoTokenizer, WhisperFeatureExtractor

        # Suppress the harmless warning about model_type mismatch when loading
        # tokenizer for custom model types not registered in transformers
        # (warning is emitted via logging, not warnings module)
        prev_verbosity = transformers.logging.get_verbosity()
        transformers.logging.set_verbosity_error()
        try:
            model._tokenizer = AutoTokenizer.from_pretrained(
                str(model_path), trust_remote_code=True
            )
            model._feature_extractor = WhisperFeatureExtractor.from_pretrained(
                str(model_path)
            )
        finally:
            transformers.logging.set_verbosity(prev_verbosity)

        if not hasattr(model.config, "model_repo") or model.config.model_repo is None:
            try:
                index = model_path.parts.index("hub")
                model.config.model_repo = (
                    model_path.parts[index + 1]
                    .replace("models--", "")
                    .replace("--", "/")
                )
            except (ValueError, IndexError):
                model.config.model_repo = str(model_path)

        return model

    def _preprocess_audio(
        self,
        audio: Union[str, mx.array, np.ndarray, List[Union[str, mx.array, np.ndarray]]],
    ) -> Tuple[mx.array, mx.array, int]:
        """Preprocess audio for the model."""
        from mlx_audio.stt.utils import load_audio

        audio_input = audio[0] if isinstance(audio, list) else audio

        if isinstance(audio_input, str):
            audio_input = load_audio(audio_input)

        audio_np = (
            np.array(audio_input) if isinstance(audio_input, mx.array) else audio_input
        )

        audio_inputs = self._feature_extractor(
            audio_np,
            sampling_rate=16000,
            return_attention_mask=True,
            truncation=False,
            padding=True,
            return_tensors="np",
        )
        input_features = mx.array(audio_inputs["input_features"])
        feature_attention_mask = mx.array(audio_inputs["attention_mask"])

        audio_lengths = feature_attention_mask.sum(axis=-1)
        aftercnn_lens = _get_feat_extract_output_lengths(audio_lengths)
        num_audio_tokens = int(aftercnn_lens[0].item())

        return input_features, feature_attention_mask, num_audio_tokens

    def _build_prompt(
        self, num_audio_tokens: int, language: str = "English"
    ) -> mx.array:
        """Build prompt with audio tokens."""
        supported = self.config.support_languages or []
        supported_lower = {lang.lower(): lang for lang in supported}

        # Match language (case-insensitive) against supported languages
        lang_name = supported_lower.get(language.lower(), language)

        prompt = (
            f"<|im_start|>system\n<|im_end|>\n"
            f"<|im_start|>user\n<|audio_start|>{'<|audio_pad|>' * num_audio_tokens}<|audio_end|><|im_end|>\n"
            f"<|im_start|>assistant\nlanguage {lang_name}<asr_text>"
        )

        input_ids = self._tokenizer.encode(prompt, return_tensors="np")
        return mx.array(input_ids)

    def stream_generate(
        self,
        audio: Union[str, mx.array, np.ndarray, List[Union[str, mx.array, np.ndarray]]],
        *,
        max_tokens: int = 8192,
        sampler: Optional[Callable[[mx.array], mx.array]] = None,
        logits_processors: Optional[List[Callable]] = None,
        language: str = "English",
        prefill_step_size: int = 2048,
        verbose: bool = False,
    ) -> Generator[Tuple[mx.array, mx.array], None, None]:
        """Stream generate tokens from audio using mlx_lm generate_step."""
        from mlx_lm.generate import generate_step

        if not hasattr(self, "_tokenizer") or not hasattr(self, "_feature_extractor"):
            raise RuntimeError(
                "Tokenizer/FeatureExtractor not initialized. Call post_load_hook first."
            )

        input_features, feature_attention_mask, num_audio_tokens = (
            self._preprocess_audio(audio)
        )
        input_ids = self._build_prompt(num_audio_tokens, language)
        eos_token_ids = [151645, 151643]

        # Step 1: Encode audio features
        with tqdm(
            total=1, desc="Encoding audio", disable=not verbose, leave=False
        ) as pbar:
            audio_features = self.get_audio_features(
                input_features, feature_attention_mask
            )
            mx.eval(audio_features)
            pbar.update(1)

        # Free input features
        del input_features, feature_attention_mask
        mx.clear_cache()

        # Step 2: Build input embeddings with audio merged
        with tqdm(
            total=1, desc="Building embeddings", disable=not verbose, leave=False
        ) as pbar:
            inputs_embeds = self._build_inputs_embeds(input_ids, audio_features)
            mx.eval(inputs_embeds)
            pbar.update(1)

        # Free audio features
        del audio_features
        mx.clear_cache()

        # Remove batch dim for generate_step
        input_embeddings = inputs_embeds[0]
        prompt = input_ids[0] if input_ids.ndim > 1 else input_ids

        # Free the batched version
        del inputs_embeds
        mx.clear_cache()

        # Progress bars
        prefill_pbar = None
        gen_pbar = None

        if verbose:
            prefill_pbar = tqdm(total=1, desc="Prefilling", unit="tok", leave=False)

        def prefill_progress(processed: int, total: int):
            nonlocal gen_pbar
            if prefill_pbar is not None:
                if prefill_pbar.total != total:
                    prefill_pbar.total = total
                    prefill_pbar.refresh()
                prefill_pbar.n = processed
                prefill_pbar.refresh()
                if processed >= total:
                    prefill_pbar.close()
                    if gen_pbar is None and verbose:
                        gen_pbar = tqdm(
                            total=max_tokens, desc="Generating", unit="tok", leave=False
                        )

        # Use generate_step from mlx_lm (handles chunked prefill automatically)
        for token, logprobs in generate_step(
            prompt=prompt,
            input_embeddings=input_embeddings,
            model=self,
            max_tokens=max_tokens,
            sampler=sampler,
            logits_processors=logits_processors,
            prefill_step_size=prefill_step_size,
            prompt_progress_callback=prefill_progress if verbose else None,
        ):
            if gen_pbar is not None:
                gen_pbar.update(1)

            if token in eos_token_ids:
                break

            yield token, logprobs

        if gen_pbar is not None:
            gen_pbar.close()

    def _generate_single_chunk(
        self,
        audio_chunk: np.ndarray,
        *,
        max_tokens: int = 8192,
        sampler: Optional[Callable] = None,
        logits_processors: Optional[List[Callable]] = None,
        language: str = "English",
        prefill_step_size: int = 2048,
        verbose: bool = False,
    ) -> Tuple[str, int, int]:
        """Generate transcription for a single audio chunk.

        Returns:
            Tuple of (text, prompt_tokens, generation_tokens).
        """
        generated_tokens = []
        prompt_tokens = 0

        for token, _ in self.stream_generate(
            audio_chunk,
            max_tokens=max_tokens,
            sampler=sampler,
            logits_processors=logits_processors,
            language=language,
            prefill_step_size=prefill_step_size,
            verbose=verbose,
        ):
            if prompt_tokens == 0:
                # Get prompt tokens from first iteration
                _, _, num_audio_tokens = self._preprocess_audio(audio_chunk)
                input_ids = self._build_prompt(num_audio_tokens, language)
                prompt_tokens = input_ids.shape[1]
            generated_tokens.append(int(token))

        text = self._tokenizer.decode(generated_tokens, skip_special_tokens=True)
        return text, prompt_tokens, len(generated_tokens)

    def generate(
        self,
        audio: Union[str, mx.array, np.ndarray, List[Union[str, mx.array, np.ndarray]]],
        *,
        max_tokens: int = 8192,
        temperature: float = 0.0,
        top_p: float = 1.0,
        top_k: int = 0,
        min_p: float = 0.0,
        min_tokens_to_keep: int = 1,
        repetition_penalty: Optional[float] = None,
        repetition_context_size: int = 100,
        language: str = "English",
        prefill_step_size: int = 2048,
        chunk_duration: float = 1200.0,
        min_chunk_duration: float = 1.0,
        verbose: bool = False,
        stream: bool = False,
        **kwargs,
    ) -> Union[STTOutput, Generator[str, None, None]]:
        """Generate transcription from audio.

        Automatically chunks long audio and processes sequentially.

        Args:
            chunk_duration: Maximum chunk duration in seconds (default: 1200 = 20 min).
            min_chunk_duration: Minimum chunk duration in seconds (default: 1.0).
            stream: If True, return a generator that yields tokens as they are generated.
        """
        # If streaming requested, delegate to stream_transcribe
        if stream:
            return self.stream_transcribe(
                audio,
                max_tokens=max_tokens,
                temperature=temperature,
                top_p=top_p,
                top_k=top_k,
                min_p=min_p,
                min_tokens_to_keep=min_tokens_to_keep,
                repetition_penalty=repetition_penalty,
                repetition_context_size=repetition_context_size,
                language=language,
                prefill_step_size=prefill_step_size,
                chunk_duration=chunk_duration,
                min_chunk_duration=min_chunk_duration,
                verbose=verbose,
            )

        from mlx_lm.sample_utils import make_logits_processors, make_sampler

        from mlx_audio.stt.utils import load_audio

        del kwargs

        start_time = time.time()

        if not hasattr(self, "_tokenizer") or not hasattr(self, "_feature_extractor"):
            raise RuntimeError(
                "Tokenizer/FeatureExtractor not initialized. Call post_load_hook first."
            )

        # Load audio
        audio_input = audio[0] if isinstance(audio, list) else audio
        if isinstance(audio_input, str):
            audio_input = load_audio(audio_input)
        audio_np = (
            np.array(audio_input) if isinstance(audio_input, mx.array) else audio_input
        )

        # Split into chunks if needed
        chunks = split_audio_into_chunks(
            audio_np,
            sr=self.sample_rate,
            chunk_duration=chunk_duration,
            min_chunk_duration=min_chunk_duration,
        )

        sampler = make_sampler(
            temperature,
            top_p,
            min_p,
            min_tokens_to_keep=min_tokens_to_keep,
            top_k=top_k,
        )

        logits_processors = (
            make_logits_processors(
                repetition_penalty=repetition_penalty,
                repetition_context_size=repetition_context_size,
            )
            if repetition_penalty
            else None
        )

        # Process chunks
        all_texts = []
        segments = []
        total_prompt_tokens = 0
        total_generation_tokens = 0

        chunk_iter = tqdm(
            chunks, desc="Processing chunks", disable=not verbose or len(chunks) == 1
        )
        for chunk_audio, offset_sec in chunk_iter:
            actual_chunk_duration = len(chunk_audio) / self.sample_rate

            text, prompt_toks, gen_toks = self._generate_single_chunk(
                chunk_audio,
                max_tokens=max_tokens,
                sampler=sampler,
                logits_processors=logits_processors,
                language=language,
                prefill_step_size=prefill_step_size,
                verbose=verbose
                and len(chunks) == 1,  # Only show inner progress for single chunk
            )
            all_texts.append(text)
            total_prompt_tokens += prompt_toks
            total_generation_tokens += gen_toks

            # Create segment for this chunk
            segments.append(
                {
                    "text": text,
                    "start": offset_sec,
                    "end": offset_sec + actual_chunk_duration,
                }
            )

            # Clear cache between chunks
            mx.clear_cache()

        end_time = time.time()

        # Combine transcriptions
        full_text = " ".join(all_texts)

        return STTOutput(
            text=full_text,
            segments=segments,
            prompt_tokens=total_prompt_tokens,
            generation_tokens=total_generation_tokens,
            total_tokens=total_prompt_tokens + total_generation_tokens,
            total_time=end_time - start_time,
            prompt_tps=(
                total_prompt_tokens / (end_time - start_time)
                if end_time > start_time
                else 0
            ),
            generation_tps=(
                total_generation_tokens / (end_time - start_time)
                if end_time > start_time
                else 0
            ),
        )

    def stream_transcribe(
        self,
        audio: Union[str, mx.array, np.ndarray, List[Union[str, mx.array, np.ndarray]]],
        *,
        max_tokens: int = 8192,
        temperature: float = 0.0,
        top_p: float = 1.0,
        top_k: int = 0,
        min_p: float = 0.0,
        min_tokens_to_keep: int = 1,
        repetition_penalty: Optional[float] = None,
        repetition_context_size: int = 100,
        language: str = "English",
        prefill_step_size: int = 2048,
        chunk_duration: float = 1200.0,
        min_chunk_duration: float = 1.0,
        verbose: bool = False,
    ) -> Generator[StreamingResult, None, None]:
        """Stream transcription token-by-token from audio.

        Automatically chunks long audio and streams tokens from each chunk sequentially.

        Yields:
            StreamingResult objects with text, timing, and status information.
        """
        from mlx_lm.sample_utils import make_logits_processors, make_sampler

        from mlx_audio.stt.utils import load_audio

        if not hasattr(self, "_tokenizer") or not hasattr(self, "_feature_extractor"):
            raise RuntimeError(
                "Tokenizer/FeatureExtractor not initialized. Call post_load_hook first."
            )

        # Load audio
        audio_input = audio[0] if isinstance(audio, list) else audio
        if isinstance(audio_input, str):
            audio_input = load_audio(audio_input)
        audio_np = (
            np.array(audio_input) if isinstance(audio_input, mx.array) else audio_input
        )

        # Calculate total audio duration
        total_duration = len(audio_np) / self.sample_rate

        # Split into chunks if needed
        chunks = split_audio_into_chunks(
            audio_np,
            sr=self.sample_rate,
            chunk_duration=chunk_duration,
            min_chunk_duration=min_chunk_duration,
        )

        sampler = make_sampler(
            temperature,
            top_p,
            min_p,
            min_tokens_to_keep=min_tokens_to_keep,
            top_k=top_k,
        )

        logits_processors = (
            make_logits_processors(
                repetition_penalty=repetition_penalty,
                repetition_context_size=repetition_context_size,
            )
            if repetition_penalty
            else None
        )

        # Normalize language code for output
        lang_code = language[:2].lower() if language else "en"

        # Process each chunk and stream tokens
        chunk_iter = tqdm(
            enumerate(chunks),
            total=len(chunks),
            desc="Processing chunks",
            disable=not verbose or len(chunks) == 1,
        )
        for chunk_idx, (chunk_audio, offset_sec) in chunk_iter:
            actual_chunk_duration = len(chunk_audio) / self.sample_rate
            is_last_chunk = chunk_idx == len(chunks) - 1
            token_count = 0

            for token, _ in self.stream_generate(
                chunk_audio,
                max_tokens=max_tokens,
                sampler=sampler,
                logits_processors=logits_processors,
                language=language,
                prefill_step_size=prefill_step_size,
                verbose=verbose and len(chunks) == 1,
            ):
                text = self._tokenizer.decode([int(token)])

                # Estimate timing based on token position within chunk
                # This is approximate since we don't have word-level alignment
                prev_progress = token_count / max(max_tokens, 1)
                token_count += 1
                curr_progress = min(token_count / max(max_tokens, 1), 1.0)

                estimated_start = offset_sec + (actual_chunk_duration * prev_progress)
                estimated_end = offset_sec + (actual_chunk_duration * curr_progress)

                yield StreamingResult(
                    text=text,
                    is_final=False,
                    start_time=estimated_start,
                    end_time=estimated_end,
                    language=lang_code,
                )

            # Yield final result for this chunk
            yield StreamingResult(
                text="",
                is_final=is_last_chunk,
                start_time=offset_sec,
                end_time=offset_sec + actual_chunk_duration,
                language=lang_code,
            )

            # Clear cache between chunks
            mx.clear_cache()


class Model:

    _FORCED_ALIGNER_TYPE = "qwen3_forced_aligner"
    _FORCED_ALIGNER_MAX_CLASSES = 10000

    def __init__(self, config):
        from .qwen3_forced_aligner import ForcedAlignerConfig, ForcedAlignerModel

        is_aligner = (
            isinstance(config, ForcedAlignerConfig)
            or getattr(config, "model_type", "") == self._FORCED_ALIGNER_TYPE
        )

        self._model = (
            ForcedAlignerModel(config) if is_aligner else Qwen3ASRModel(config)
        )
        self.config = self._model.config

    def __getattr__(self, name):
        return getattr(self._model, name)

    def __call__(self, *args, **kwargs):
        return self._model(*args, **kwargs)

    def parameters(self):
        return self._model.parameters()

    def load_weights(self, weights, strict=False):
        return self._model.load_weights(weights, strict=strict)

    def eval(self):
        return self._model.eval()

    @classmethod
    def _is_forced_aligner_weights(cls, weights: Dict[str, mx.array]) -> bool:
        """Check if weights are for ForcedAligner based on lm_head output size."""
        for key, value in weights.items():
            if "lm_head" in key and "weight" in key:
                return value.shape[0] < cls._FORCED_ALIGNER_MAX_CLASSES
        return False

    @classmethod
    def sanitize(cls, weights: Dict[str, mx.array]) -> Dict[str, mx.array]:
        if cls._is_forced_aligner_weights(weights):
            from .qwen3_forced_aligner import ForcedAlignerModel

            return ForcedAlignerModel.sanitize(weights)
        return Qwen3ASRModel.sanitize(weights)

    @classmethod
    def post_load_hook(cls, model: "Model", model_path: Path) -> "Model":
        internal_cls = type(model._model)
        if hasattr(internal_cls, "post_load_hook"):
            model._model = internal_cls.post_load_hook(model._model, model_path)
            model.config = model._model.config
        return model



from .models.mossformer2_se import (
    MossFormer2SE,
    MossFormer2SEConfig,
    MossFormer2SEModel,
)
from .models.sam_audio import (
    Batch,
    SAMAudio,
    SAMAudioConfig,
    SAMAudioProcessor,
    SeparationResult,
    save_audio,
)
from .voice_pipeline import VoicePipeline

__all__ = [
    "SAMAudio",
    "SAMAudioProcessor",
    "SeparationResult",
    "Batch",
    "save_audio",
    "SAMAudioConfig",
    "VoicePipeline",
    # MossFormer2 SE
    "MossFormer2SE",
    "MossFormer2SEConfig",
    "MossFormer2SEModel",
]
